﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using PaintDotNet;
using PaintDotNet.Effects;
using PaintDotNet.IndirectUI;
using PaintDotNet.PropertySystem;

namespace ILikePi.DiagonalLines {
    [EffectTypeHint(EffectTypeHint.Fast | EffectTypeHint.Unary)]
    [EffectCategory(EffectCategory.Effect)]
    public class DiagonalLinesEffect : PropertyBasedEffect {
        private static readonly PropertyCollection propCollection;

        static DiagonalLinesEffect() {
            ICollection<Property> props = new List<Property>();
            props.Add(new Int32Property("Interval", 10, 2, 200));
            props.Add(new Int32Property("Brush width", 1, 1, 200));
            props.Add(new DoubleProperty("Angle", 45, 0, 360));
            props.Add(new BooleanProperty("Antialias", false));
            propCollection = new PropertyCollection(props);
        }

        public DiagonalLinesEffect()
            : base("Diagonal Lines", Properties.Resources.Icon, SubmenuNames.Render /* Localized name for Render menu */, EffectFlags.Configurable) {
        }

        protected override PropertyCollection OnCreatePropertyCollection() {
            return propCollection;
        }

        protected override ControlInfo OnCreateConfigUI(PropertyCollection props) {
            ControlInfo info = CreateDefaultConfigUI(props);
            // Get an angle chooser instead of a slider
            info.SetPropertyControlType("Angle", PropertyControlType.AngleChooser);
            return info;
        }

        public override unsafe void OnRender(Rectangle[] renderRects, int startIndex, int length) {
            // Copy over the source surface
            Surface srcSurface = SrcArgs.Surface, dstSurface = DstArgs.Surface;
            foreach (Rectangle roi in renderRects) {
                for (int y = roi.Top; y < roi.Bottom; y++) {
                    // Pointers are faster
                    ColorBgra* srcColor = srcSurface.GetRowAddressUnchecked(y);
                    ColorBgra* dstColor = dstSurface.GetRowAddressUnchecked(y);
                    srcColor += roi.Left;
                    dstColor += roi.Left;
                    for (int x = 0; x < roi.Width; x++) {
                        dstColor->Bgra = srcColor->Bgra;
                        srcColor++;
                        dstColor++;
                    }
                }
            }

            double angle = Token.GetProperty<DoubleProperty>("Angle").Value;
            // If lines are horizontal, xOffset will be infinity
            if (angle % 180 != 0) {
                Rectangle selection = EnvironmentParameters.GetSelection(SrcArgs.Bounds).GetBoundsInt();
                Graphics g = DstArgs.Graphics;
                Pen pen = new Pen(EnvironmentParameters.PrimaryColor.ToColor(), Token.GetProperty<Int32Property>("Brush width").Value);
                int interval = Token.GetProperty<Int32Property>("Interval").Value;
                //     /|
                //    / |
                //   /  | h
                //  /   |
                // /a___|
                //   dx
                double dx = selection.Height / Math.Tan(angle * Math.PI / 180 /* convert to radians */);

                // Lock to make multithread safe
                lock (typeof(DiagonalLinesEffect)) {
                    if (Token.GetProperty<BooleanProperty>("Antialias").Value) {
                        g.SmoothingMode = SmoothingMode.AntiAlias;
                    } else {
                        g.SmoothingMode = SmoothingMode.None;
                    }

                    // Make sure we only draw in renderRects
                    Region region = new Region();
                    // Took me a while to figure out that a default region is infinite. Was wondering why antialiasing looked bad...
                    region.MakeEmpty();
                    foreach (Rectangle rect in renderRects) {
                        region.Union(rect);
                    }
                    g.SetClip(region, CombineMode.Replace);

                    int starti;
                    if (dx < 0) {
                        starti = (int)(selection.Left + dx);
                    } else {
                        starti = selection.Left + interval;
                    }

                    int endi;
                    if (dx < 0) {
                        endi = selection.Right;
                    } else {
                        endi = (int)(selection.Right + dx);
                    }

                    for (int i = starti; i < endi; i += interval) {
                        g.DrawLine(pen, i, selection.Top, (int)(i - dx), selection.Bottom);
                    }
                }
            }
        }
    }
}